import numpy as np

from ... import defaults
from .data import Data
from .bar import Bar

class Bars(Data):

    def __init__(self, bar_objects, **kwargs):
        Data.__init__(self, **kwargs)
        defaults.kwargs(self, kwargs)
        self.set_bar_objects(bar_objects)

    def set_bar_objects(self, bar_objects):
        if isinstance(bar_objects, Bar):
            self.bar_objects = [bar_objects]
        else:
            self.set_bar_objects_multiple(bar_objects)

    def set_bar_objects_multiple(self, bar_objects):
        if np.all([isinstance(bar_obj, Bar)
                   for bar_obj in bar_objects]):
            self.bar_objects = list(bar_objects)
        else:
            self.bad_data_objects_exception()

    def bad_data_objects_exception(self):
        message = ("When creating a bars object you must pass "
                   "in an instance of bar or an iterable of "
                   "instances of bar")
        raise TypeError(message)

defaults.load(Bars)



import numpy as np

from ... import defaults
from .data import Data

class Colorplot(Data):

    def __init__(self, *args, **kwargs):
        Data.__init__(self, **kwargs)
        defaults.kwargs(self, kwargs)
        self.process_args(args)

    def process_args(self, args):
        if len(args) == 3:
            self.x, self.y, self.z_mesh = args
        elif len(args) == 1:
            self.z_mesh = args[0]
        else:
            self.incorrect_positional_args_exception(args)

    def incorrect_positional_args_exception(self, args):
        message = (f"Colorplot takes 1 or 3 positional arguments, got {len(args)}\n"
                   "If one given, it is assumed to be the mesh data\n"
                   "If three are given, it is expected to be in the form (x, y, z)")
        raise Exception(message)

defaults.load(Colorplot)



import os
import math
from screeninfo import get_monitors

import matplotlib.pyplot as plt
import numpy as np

from .. import defaults
from .plottypes.plotlines import PlotLines
from .plottypes.plotbars import PlotBars
from .plottypes.plotpie import PlotPie
from .plottypes.plotcolorplot import PlotColorplot
from .plottypes.plotsurface import PlotSurface
from .plotutils.griddimensions import get_grid_dimensions
from .plotutils.savefigure import save_figure
from .plotutils.figuresize import maximise_figure
from ..utils.dicts import remove_none_values

class Figure():
    
    @classmethod
    def set_plot_classes(cls):
        cls.plot_classes = {"Lines": PlotLines,
                            "Bars": PlotBars,
                            "Pie": PlotPie,
                            "Colorplot": PlotColorplot,
                            "Surface": PlotSurface}
    
    def __init__(self, figures_obj, data_objects, plot_index, **kwargs):
        defaults.kwargs(self, kwargs)
        self.figures_obj = figures_obj
        self.plot_index = plot_index
        self.initialise_data_objects(data_objects)
        self.set_grid_size()
        self.process_light_or_dark()

    def process_light_or_dark(self):
        if self.dark:
            plt.style.use('dark_background')

    def initialise_data_objects(self, data_objects):
        self.data_objects = data_objects
        self.count = len(data_objects)
        if self.figures_obj.universal_legend:
            self.count += 1

    def set_grid_size(self):
        aspect_ratio = self.figures_obj.aspect_ratio
        self.rows, self.columns = get_grid_dimensions(self.count, aspect_ratio)

    def create_figure(self):
        self.initialise_figure()
        self.create_plots()
        self.add_figure_peripherals()
        self.output_figure()

    def initialise_figure(self):
        self.set_figure()
        self.create_axes()
        self.remove_extra_axes()

    def set_figure(self):
        self.fig = plt.figure(constrained_layout=True, dpi=self.dpi)
        self.fig.set_constrained_layout_pads(w_pad=self.w_pad, h_pad=self.h_pad,
                                             hspace=self.hspace, wspace=self.wspace)

    def create_axes(self):
        self.axes = [self.get_axis(index, data_obj)
                     for index, data_obj in enumerate(self.data_objects)]

    def get_axis(self, index, data_obj):
        axis = self.fig.add_subplot(self.rows, self.columns, index + 1,
                                    projection=data_obj.projection)
        return axis
    
    def remove_extra_axes(self):
        extra_axes_count = len(self.axes) - self.count
        for ax, _ in zip(self.axes[::-1], range(extra_axes_count)):
            ax.remove()

    def add_figure_peripherals(self):
        self.set_suptitle_fontdict()
        self.set_suptitle()
        self.set_universal_legend()
        self.set_figure_size()

    def set_suptitle_fontdict(self):
        kwargs = {"fontname": self.suptitle_fontname,
                  "fontsize": self.suptitle_fontsize,
                  "color": self.suptitle_color,
                  "verticalalignment": self.suptitle_verticalalignment,
                  "horizontalalignment": self.suptitle_horizontalalignment}
        self.suptitle_fontdict = remove_none_values(kwargs)

    def set_suptitle(self):
        if self.figures_obj.suptitle is not None:
            self.fig.suptitle(f"{self.figures_obj.suptitle}",
                              **self.suptitle_fontdict, loc=self.suptitle_loc,
                              y=self.suptitle_y, pad=self.suptitle_pad)

    def set_universal_legend(self):
        if self.figures_obj.universal_legend:
            self.do_universal_legend()

    def do_universal_legend(self):
        ax = self.axes[-1]
        for data_obj in self.data_objects[0].data_objects:
            ax.plot(1, 1, label=data_obj.label, color=data_obj.colour)
        ax.legend(loc="center", borderpad=2, labelspacing=1)
        ax.axis("off")

    def set_figure_size(self):
        self.update_size()
        self.set_figure_size_pixels()

    def update_size(self):
        self.set_maximised_figure_size()
        self.set_figure_inches()
        self.maximise_figure()

    def set_maximised_figure_size(self):
        if self.maximise:
            monitor = get_monitors()[0]
            figure_size = [monitor.width_mm / 25.4, monitor.height_mm / 25.4]
            self.figure_size = figure_size

    def maximise_figure(self):
        if self.maximise:
            maximise_figure()

    def set_figure_inches(self):
        if self.figure_size is not None:
            self.fig.set_size_inches(self.figure_size)

    def set_figure_size_pixels(self):
        self.figure_size_pixels = self.fig.get_size_inches()*self.fig.dpi
        self.figure_size_pixels = [int(value) for value in self.figure_size_pixels]

    def create_plots(self):
        self.plot_objects = [self.create_plot_obj(ax, data_obj)
                             for ax, data_obj in zip(self.axes, self.data_objects)]

    def create_plot_obj(self, ax, data_obj):
        plot_class = self.plot_classes[data_obj.__class__.__name__]
        plot_obj = plot_class(self, ax, data_obj)
        plot_obj.create_plot()
        return plot_obj
    
    def output_figure(self):
        if self.figures_obj.output == "Show":
            self.show_figure()
        elif self.figures_obj.output == "Save":
            save_figure(self)

    def show_figure(self):
        plt.show()
        plt.close()

    def set_animation_axis_limits(self):
        for data_obj in self.data_objects:
            data_obj.set_animation_axis_limits()

    def get_frame_count(self):
        frame_count = self.data_objects[0].get_frame_count()
        return frame_count

    def set_data_value(self, index):
        for data_obj, data_values in zip(self.data_objects, self.all_data_values):
            data_value = data_values[index]
            data_obj.set_data_value(data_value)

defaults.load(Figure)



import matplotlib.pyplot as plt
import numpy as np

from ... import defaults
from ..plottypes.plot import Plot

class PlotSurface(Plot):

    @classmethod
    def set_function_dict(cls):
        cls.function_dict = {"plot_surface": cls.plot_surface,
                             "plot_wireframe": cls.plot_wireframe,
                             "contour": cls.plot_contour}

    def __init__(self, figure_obj, ax, data_obj, **kwargs):
        Plot.__init__(self, figure_obj, ax, data_obj, **kwargs)
        defaults.kwargs(self, kwargs)

    def plot_data(self):
        function_type = self.get_plot_function()
        plot_function = getattr(self.ax, self.data_obj.plot_type)
        function_type(self, plot_function, self.data_obj)
        self.set_z_limits()
        self.set_axes()

    def get_plot_function(self):
        plot_type = self.data_obj.plot_type
        plot_function = self.function_dict[plot_type]
        return plot_function

    def plot_surface(self, plot_function, data_obj):
        plot_function(data_obj.x_values,
                      data_obj.y_values,
                      data_obj.z_values,
                      color=data_obj.color,
                      cmap=data_obj.cmap,
                      facecolors=data_obj.facecolors,
                      norm=data_obj.norm,
                      vmin=data_obj.vmin,
                      vmax=data_obj.vmax,
                      shade=data_obj.shade,
                      lightsource=data_obj.lightsource)
    
    def plot_wireframe(self, plot_function, data_obj):
        plot_function(data_obj.x_values,
                      data_obj.y_values,
                      data_obj.z_values,
                      rcount=data_obj.rcount,
                      ccount=data_obj.ccount,
                      colors=data_obj.colors,
                      cmap=data_obj.cmap,
                      facecolors=data_obj.facecolors,
                      norm=data_obj.norm,
                      segments=data_obj.segments,
                      linewidths=data_obj.linewidths,
                      antialiased=data_obj.antialiased,
                      zorder=data_obj.zorder,
                      edgecolors=data_obj.edgecolors,
                      linestyles=data_obj.linestyles,
                      capstyle=data_obj.capstyle,
                      joinstyle=data_obj.joinstyle,
                      offsets=data_obj.offsets,
                      offset_transform=data_obj.offset_transform,
                      hatch=data_obj.hatch)

    def plot_contour(self, plot_function, data_obj):
        plot_function(data_obj.x_values,
                      data_obj.y_values,
                      data_obj.z_values,
                      extend3d=data_obj.extend3d,
                      stride=data_obj.stride,
                      zdir=data_obj.zdir,
                      offset=data_obj.offset,
                      levels=data_obj.levels,
                      corner_mask=data_obj.corner_mask,
                      colors=data_obj.colors,
                      cmap=data_obj.cmap,
                      norm=data_obj.norm,
                      vmin=data_obj.vmin,
                      vmax=data_obj.vmax,
                      origin=data_obj.origin,
                      extent=data_obj.extent,
                      locator=data_obj.locator,
                      extend=data_obj.extend)
    
    def set_z_limits(self):
        if self.data_obj.z_limits is not None:
            bottom, top = self.data_obj.z_limits
            self.ax.set_zlim(bottom=bottom, top=top)

    def set_axes(self):
        if not self.data_obj.axes:
            self.ax.axis("off")

defaults.load(PlotSurface)



import matplotlib.pyplot as plt
import numpy as np

from ... import defaults
from ..plottypes.plot import Plot

class PlotPie(Plot):

    def __init__(self, figure_obj, ax, pie_obj, **kwargs):
        Plot.__init__(self, figure_obj, ax, pie_obj, **kwargs)
        self.pie_obj = pie_obj
        defaults.kwargs(self, kwargs)

    def plot_data(self):
        pie_obj = self.pie_obj
        self.ax.pie(pie_obj.values,
                    labels=pie_obj.labels,
                    explode=pie_obj.explode,
                    colors=pie_obj.colors,
                    hatch=pie_obj.hatch,
                    autopct=pie_obj.autopct,
                    pctdistance=pie_obj.pctdistance,
                    labeldistance=pie_obj.labeldistance,
                    shadow=pie_obj.shadow,
                    startangle=pie_obj.startangle,
                    radius=pie_obj.radius,
                    counterclock=pie_obj.counterclock,
                    wedgeprops=pie_obj.wedgeprops,
                    textprops=pie_obj.textprops,
                    center=pie_obj.center,
                    frame=pie_obj.frame,
                    rotatelabels=pie_obj.rotatelabels,
                    normalize=pie_obj.normalize)

defaults.load(PlotPie)



import math

class GridDimensions():

    def __init__(self, count, aspect_ratio):
        self.count = count
        self.aspect_ratio = aspect_ratio
        self.set_grid_size()

    def set_grid_size(self):
        self.set_row_and_column_sizes()
        self.set_row_column_pairs()
        self.select_row_column_pair()
        self.dimensions = [self.rows, self.columns]

    def set_row_and_column_sizes(self):
        self.set_row_sizes()
        self.set_column_sizes()

    def set_row_sizes(self):
        row_size = math.sqrt(self.count / self.aspect_ratio)
        self.row_small = math.floor(row_size)
        self.row_large = math.ceil(row_size)

    def set_column_sizes(self):
        column_size = math.sqrt(self.count * self.aspect_ratio)
        self.column_small = math.floor(column_size)
        self.column_large = math.ceil(column_size)

    def set_row_column_pairs(self):
        self.set_pairs()
        self.row_column_pairs = {pair: {} for pair in self.pairs}
        self.populate_row_column_pairs()

    def set_pairs(self):
        self.pairs = [(self.row_small, self.column_small),
                      (self.row_small, self.column_large),
                      (self.row_large, self.column_small),
                      (self.row_large, self.column_large)]

    def populate_row_column_pairs(self):
        for pair in self.row_column_pairs.keys():
            self.row_column_pairs[pair] = self.get_row_column_pair_data(pair)

    def get_row_column_pair_data(self, pair):
        size = pair[0] * pair[1]
        aspect_ratio = self.get_aspect_ratio(pair)
        pair_data_dict = {"Size": size, "Aspect Ratio": aspect_ratio}
        return pair_data_dict

    def get_aspect_ratio(self, pair):
        if pair[0] != 0 and pair[1] != 0:
            aspect_ratio = pair[0] / pair[1]
        else:
            aspect_ratio = None
        return aspect_ratio

    def select_row_column_pair(self):
        if len(self.row_column_pairs) == 1:
            self.exact_ratio_pair()
        else:
            self.non_exact_ratio_pairs()

    def exact_ratio_pair(self):
        self.rows, self.columns = self.pairs[0]

    def non_exact_ratio_pairs(self):
        is_grid_big_enough = self.get_is_grid_big_enough()
        row_column_pair_functions = self.get_row_column_pair_functions()
        row_column_pair_function, *args = row_column_pair_functions[is_grid_big_enough]
        row_column_pair_function(*args)
        self.rows, self.columns = self.best_pair

    def get_is_grid_big_enough(self):
        is_grid_big_enough = tuple([row_column_pair["Size"] >= self.count
                                    for row_column_pair in self.row_column_pairs.values()])
        return is_grid_big_enough

    def get_row_column_pair_functions(self):
        row_column_pair_functions = {(True, True, True, True): (self.set_best_pair, 0),
                                     (False, True, True, True): (self.middle_pair_compare,),
                                     (False, True, False, True): (self.set_best_pair, 1),
                                     (False, False, True, True): (self.set_best_pair, 2),
                                     (False, False, False, True): (self.set_best_pair, 3)}
        return row_column_pair_functions

    def set_best_pair(self, pair_index):
        self.best_pair = self.pairs[pair_index]

    def middle_pair_compare(self):
        self.set_aspect_ratio_scores()
        self.compare_aspect_ratio_scores()

    def set_aspect_ratio_scores(self):
        aspect_ratio_1 = self.row_column_pairs[self.pairs[1]]["Aspect Ratio"]
        aspect_ratio_2 = self.row_column_pairs[self.pairs[2]]["Aspect Ratio"]
        self.aspect_ratio_score_1 = self.get_aspect_ratio_score(aspect_ratio_1)
        self.aspect_ratio_score_2 = self.get_aspect_ratio_score(aspect_ratio_2)

    def get_aspect_ratio_score(self, aspect_ratio):
        if aspect_ratio is not None:
            score = max(aspect_ratio / self.aspect_ratio,
                        self.aspect_ratio / aspect_ratio)
        else:
            score = None
        return score

    def compare_aspect_ratio_scores(self):
        if self.aspect_ratio_score_1 < self.aspect_ratio_score_2:
            self.set_best_pair(1)
        else:
            self.set_best_pair(2)

def get_grid_dimensions(count, aspect_ratio):
    grid_dimensions_obj = GridDimensions(count, aspect_ratio)
    return grid_dimensions_obj.dimensions



from ... import defaults

class Line():

    def __init__(self, x_values, y_values, **kwargs):
        defaults.kwargs(self, **kwargs)
        self.x_values = x_values
        self.y_values = y_values

defaults.load(Line)



import math

import numpy as np

def get_group_size(group_size, object_list):
    if group_size is None:
        group_size = len(object_list)
    return group_size

def get_group_indexes(length, group_size):
    group_size, group_count = get_group_data(length, group_size)
    end_point_indexes = get_end_point_indexes(length, group_size, group_count)
    group_indexes = construct_group_indexes(length, group_count, end_point_indexes)
    return group_indexes

def get_group_data(length, group_size):
    group_size = get_modified_group_size(length, group_size)
    group_count = math.floor(length/group_size)
    return group_size, group_count

def get_modified_group_size(list_size, average_size):
    if list_size < average_size:
        average_size = list_size
    return average_size

def get_end_point_indexes(length, group_size, group_count):
    real_group_size = length/group_count
    real_group_end_points = np.arange(0, group_count + 1) * real_group_size
    real_group_end_points = np.round(real_group_end_points, 5)
    end_point_indexes = np.ceil(real_group_end_points)
    return end_point_indexes

def construct_group_indexes(length, group_count, end_point_indexes):
    group_indexes = [np.arange(end_point_indexes[group_number],
                               end_point_indexes[group_number + 1]).astype('int')
                     for group_number in range(group_count)]
    return group_indexes

def get_group_indexes_fill(length, group_size):
    full_groups = math.floor(length / group_size)
    leftover = length - group_size*full_groups
    group_indexes = [np.arange(group_index*group_size, (group_index + 1)*group_size)
                     for group_index in range(full_groups)]
    if leftover != 0:
        group_indexes.append(np.arange(full_groups*group_size, length))
    return group_indexes




import matplotlib.pyplot as plt
import matplotlib.font_manager
import numpy as np

from ... import defaults
from ..plottypes.plot import Plot

class PlotColorplot(Plot):

    def __init__(self, figure_obj, ax, colorplot_obj, **kwargs):
        Plot.__init__(self, figure_obj, ax, colorplot_obj, **kwargs)
        self.colorplot_obj = colorplot_obj
        defaults.kwargs(self, kwargs)

    def plot_data(self):
        colorplot_obj = self.colorplot_obj
        x_and_y = self.get_x_and_y()
        self.ax.pcolormesh(*x_and_y, colorplot_obj.z_mesh,
                           cmap=colorplot_obj.cmap,
                           norm=colorplot_obj.norm,
                           vmin=colorplot_obj.vmin,
                           vmax=colorplot_obj.vmax,
                           edgecolors=colorplot_obj.edgecolors,
                           alpha=colorplot_obj.alpha,
                           shading=colorplot_obj.shading,
                           snap=colorplot_obj.snap,
                           rasterized=colorplot_obj.rasterized,
                           agg_filter=colorplot_obj.agg_filter,
                           animated=colorplot_obj.animated,
                           antialiased=colorplot_obj.antialiased,
                           capstyle=colorplot_obj.capstyle,
                           clip_box=colorplot_obj.clip_box,
                           clip_on=colorplot_obj.clip_on,
                           clip_path=colorplot_obj.clip_path,
                           color=colorplot_obj.color,
                           edgecolor=colorplot_obj.edgecolor,
                           facecolor=colorplot_obj.facecolor,
                           gid=colorplot_obj.gid,
                           hatch=colorplot_obj.hatch,
                           joinstyle=colorplot_obj.joinstyle,
                           label=colorplot_obj.label,
                           linestyle=colorplot_obj.linestyle,
                           linewidth=colorplot_obj.linewidth,
                           mouseover=colorplot_obj.mouseover,
                           offsets=colorplot_obj.offsets,
                           path_effects=colorplot_obj.path_effects,
                           visible=colorplot_obj.visible,
                           url=colorplot_obj.url,
                           zorder=colorplot_obj.zorder)

    def get_x_and_y(self):
        if ((self.colorplot_obj.x is not None) and
            (self.colorplot_obj.y is not None)):
            return (self.colorplot_obj.x, self.colorplot_obj.y)
        else:
            return ()

defaults.load(PlotColorplot)



import matplotlib.pyplot as plt
import numpy as np

from ... import defaults
from ..plottypes.plot import Plot

class PlotBars(Plot):

    def __init__(self, figure_obj, ax, bars_obj, **kwargs):
        Plot.__init__(self, figure_obj, ax, bars_obj, **kwargs)
        self.bars_obj = bars_obj
        self.initialise_from_bars_obj()
        defaults.kwargs(self, kwargs)

    def initialise_from_bars_obj(self):
        self.bar_objects = self.bars_obj.bar_objects
        self.between_group_spacing = self.bars_obj.between_group_spacing
        self.within_group_spacing = self.bars_obj.within_group_spacing

    def plot_data(self):
        self.plot_bars()
        self.set_tick_labels()

    def plot_bars(self):
        self.preprocess_bars()
        for index, bar_obj in enumerate(self.bars_obj.bar_objects):
            self.plot_bar_obj(index, bar_obj)

    def preprocess_bars(self):
        self.group_width = 1 - self.between_group_spacing
        self.set_bar_proportions()
        self.x_offset = (self.between_group_spacing + self.bar_width - 1) / 2
        self.set_x_axis_dictionary()
        
    def set_bar_proportions(self):
        if len(self.bar_objects) == 1:
            self.set_bar_proportions_single()
        else:
            self.set_proportions_multiple()

    def set_bar_proportions_single(self):
        self.bar_width = self.group_width
        self.bar_space = self.group_width

    def set_proportions_multiple(self):
        n = len(self.bar_objects)
        self.bar_width = self.group_width / (n + (n-1) * self.within_group_spacing)
        self.bar_space = (self.group_width - n * self.bar_width) / (n - 1)

    def set_x_axis_dictionary(self):
        self.x_axis_dictionary = {}
        for bar_obj in self.bar_objects:
            for x_value in bar_obj.x_values:
                self.add_to_x_axis_dictionary(x_value)

    def add_to_x_axis_dictionary(self, x_value):
        if x_value not in self.x_axis_dictionary:
            new_dict_entry = {x_value: len(self.x_axis_dictionary)}
            self.x_axis_dictionary.update(new_dict_entry)

    def get_x_values(self, index, bar_obj):
        x_values = np.array([self.x_axis_dictionary[x_value]
                             for x_value in bar_obj.x_values])
        x_values = x_values + self.x_offset + (self.bar_width + self.bar_space) * index
        return x_values

    def plot_bar_obj(self, index, bar_obj):
        bars_obj = self.bars_obj
        x_values = self.get_x_values(index, bar_obj)
        self.ax.bar(x_values,
                    bar_obj.y_values,
                    width=self.bar_width,
                    bottom=bars_obj.bottom,
                    color=bar_obj.color,
                    edgecolor=bar_obj.edgecolor,
                    linewidth=bar_obj.linewidth,
                    tick_label=bar_obj.tick_label,
                    label=bar_obj.label,
                    xerr=bar_obj.xerr,
                    yerr=bar_obj.yerr,
                    ecolor=bar_obj.ecolor,
                    capsize=bar_obj.capsize,
                    log=bars_obj.log,
                    agg_filter=bar_obj.agg_filter,
                    alpha=bar_obj.alpha,
                    angle=bar_obj.angle,
                    animated=bars_obj.animated,
                    antialiased=bar_obj.antialiased)

    def add_axis_labels(self):
        self.add_x_label()
        self.add_y_label()

defaults.load(PlotBars)



from ... import defaults

class Data():

    def __init__(self, **kwargs):
        defaults.kwargs(self, kwargs)

    def get_frame_count(self):
        raise Exception("This data type does not have animation implemented")

defaults.load(Data)



from ... import defaults

class Bar():
 
    def __init__(self, x_values, y_values, **kwargs):
        defaults.kwargs(self, kwargs)
        self.x_values = x_values
        self.y_values = y_values

defaults.load(Bar)
